==> bucketlib-fabric-1.21.3-4.2.0.7-sources.jar: de/cech12/bucketlib/client/model/UniversalBucketUnbakedModel.java <==
package de.cech12.bucketlib.client.model;

import com.google.common.collect.Maps;
import de.cech12.bucketlib.api.BucketLib;
import de.cech12.bucketlib.api.BucketLibTags;
import de.cech12.bucketlib.api.item.UniversalBucketItem;
import de.cech12.bucketlib.client.BucketLibClientMod;
import de.cech12.bucketlib.mixin.BlockModelAccessor;
import de.cech12.bucketlib.util.BucketLibUtil;
import net.fabricmc.fabric.api.client.render.fluid.v1.FluidRenderHandlerRegistry;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariantAttributes;
import net.minecraft.class_1047;
import net.minecraft.class_1058;
import net.minecraft.class_1086;
import net.minecraft.class_1087;
import net.minecraft.class_1100;
import net.minecraft.class_1309;
import net.minecraft.class_1723;
import net.minecraft.class_1799;
import net.minecraft.class_2960;
import net.minecraft.class_3611;
import net.minecraft.class_3612;
import net.minecraft.class_3665;
import net.minecraft.class_4730;
import net.minecraft.class_638;
import net.minecraft.class_777;
import net.minecraft.class_7775;
import net.minecraft.class_7923;
import net.minecraft.class_793;
import net.minecraft.class_799;
import net.minecraft.class_806;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;

/**
 * This implementation is based on net.neoforged.neoforge.client.model.DynamicFluidContainerModel.
 * Multiple changes were done to simplify the class
 */
public class UniversalBucketUnbakedModel extends class_793 implements class_1100 {

    private static final Map<class_2960, class_2960> TEXTURE_MAP = Maps.newHashMap();
    private static final class_4730 MISSING_LOWER_CONTENT_MATERIAL = new class_4730(class_1723.field_21668, getContentTexture(BucketLib.id("missing_lower_content")));

    @Nonnull
    private class_3611 fluid = class_3612.field_15906;
    @Nullable
    private class_2960 otherContent = null;

    private boolean isCracked = false;
    private boolean isLower = false;

    public UniversalBucketUnbakedModel(class_793 blockModel) {
        super(
                ((BlockModelAccessor) blockModel).bucketlib_getParentLocation(),
                blockModel.method_3433(),
                ((BlockModelAccessor) blockModel).bucketlib_getTextureMap(),
                blockModel.method_3444(),
                blockModel.method_24298(),
                blockModel.method_3443(),
                blockModel.method_3434()
        );
    }

    public UniversalBucketUnbakedModel(class_793 blockModel, @Nonnull class_3611 fluid, @Nullable class_2960 otherContent, boolean isCracked, boolean isLower) {
        this(blockModel);
        this.fluid = fluid;
        this.otherContent = otherContent;
        this.isCracked = isCracked;
        this.isLower = isLower;
    }

    /**
     * Returns a new UniversalBucketModel with the given fluid.
     */
    public UniversalBucketUnbakedModel withFluid(class_3611 newFluid, boolean isCracked) {
        return new UniversalBucketUnbakedModel(this, newFluid, null, isCracked, false);
    }

    /**
     * Returns a new UniversalBucketModel with the given other content.
     */
    public UniversalBucketUnbakedModel withOtherContent(class_2960 otherContent, boolean isCracked, boolean isLower) {
        return new UniversalBucketUnbakedModel(this, class_3612.field_15906, otherContent, isCracked, isLower);
    }

    private static class_2960 getContentTexture(class_2960 otherContentLocation) {
        class_2960 texture = TEXTURE_MAP.get(otherContentLocation);
        if (texture == null) {
            String textureLocation = String.format("item/bucket_content/%s", otherContentLocation.method_12832());
            texture = otherContentLocation.method_45136(textureLocation);
            TEXTURE_MAP.put(otherContentLocation, texture);
        }
        return texture;
    }

    public class_2960 getParentLocation() {
        return this.field_4247;
    }

    @Nonnull
    @Override
    public class_793 method_3431() {
        if (this.field_4253 == null) {
            return this;
        }
        if (BucketLibClientMod.UNIVERSAL_BUCKET_MODEL.equals(this.getParentLocation())) {
            //skip the BlockModel swap of the vanilla ItemRendering (ItemModelGenerator)
            return this.field_4253;
        }
        return this.field_4253.method_3431();
    }

    @Nonnull
    @Override
    public class_1087 method_4753(@Nonnull class_7775 modelBaker, @Nonnull Function<class_4730, class_1058> spriteGetter, @Nonnull class_3665 modelState) {
        //resolve parents to use their defined materials, too
        if (this.field_4253 == null) {
            method_62326(BucketLibClientMod::getModel);
        }
        class_4730 particleLocation = this.method_24077("particle");
        class_1058 particleSprite = null;
        if (isValid(particleLocation)) {
            particleSprite = spriteGetter.apply(particleLocation);
        }

        class_4730 baseLocation = null;
        if (this.isLower) {
            if (this.isCracked) {
                baseLocation = this.method_24077("crackedLowerBase");
            }
            if (!isValid(baseLocation)) {
                baseLocation = this.method_24077("lowerBase");
            }
        } else {
            if (this.isCracked) {
                baseLocation = this.method_24077("crackedBase");
            }
            if (!isValid(baseLocation)) {
                baseLocation = this.method_24077("base");
            }
        }

        class_4730 otherContentLocation = null;
        class_4730 fluidMaskLocation = null;
        if (this.otherContent != null) {
            otherContentLocation = new class_4730(class_1723.field_21668, getContentTexture(this.otherContent));
        } else if (this.fluid != class_3612.field_15906) {
            if (this.isCracked) {
                fluidMaskLocation = this.method_24077("crackedFluidMask");
            }
            if (!isValid(fluidMaskLocation)) {
                fluidMaskLocation = this.method_24077("fluidMask");
            }
        }

        class_1058 baseSprite = null;
        if (isValid(baseLocation)) {
            baseSprite = spriteGetter.apply(baseLocation);
            if (particleSprite == null)
                particleSprite = baseSprite;
        }
        class_1058 otherContentSprite;
        if (isValid(otherContentLocation)) {
            otherContentSprite = spriteGetter.apply(otherContentLocation);
            if (particleSprite == null)
                particleSprite = otherContentSprite;
        } else {
            //if content texture is missing - fallback to pink content texture>
            otherContentSprite = spriteGetter.apply(MISSING_LOWER_CONTENT_MATERIAL);
        }

        class_1058 fluidSprite = null;
        if (fluid != class_3612.field_15906 && FluidRenderHandlerRegistry.INSTANCE.get(fluid) != null) {
            class_1058[] sprites = Objects.requireNonNull(FluidRenderHandlerRegistry.INSTANCE.get(fluid)).getFluidSprites(null, null, fluid.method_15785());
            if (sprites.length > 0) {
                fluidSprite = sprites[0];
                if (particleSprite == null)
                    particleSprite = fluidSprite;
            }
        }

        // if the fluid is lighter than air, will manipulate the initial state to be rotated 180deg to turn it upside down
        if (fluid != class_3612.field_15906 && !fluid.method_15785().method_15767(BucketLibTags.Fluids.NO_FLIPPING) && FluidVariantAttributes.isLighterThanAir(FluidVariant.of(fluid))) {
            modelState = class_1086.field_5358;
        }

        List<class_777> quads = new ArrayList<>();

        if (baseSprite != null) {
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemElements(0, "base", baseSprite.method_45851()),
                    baseSprite, modelState));
        }
        if (isValid(otherContentLocation)) {
            assert otherContentSprite != null;
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemElements(-1, "content", otherContentSprite.method_45851()),
                    otherContentSprite, modelState));
        } else if (isValid(fluidMaskLocation) && fluid != class_3612.field_15906) {
            class_1058 templateSprite = spriteGetter.apply(fluidMaskLocation);
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemMaskElements(1, "fluid", templateSprite.method_45851()),
                    fluidSprite, modelState));
        }

        ContainedFluidOverrideHandler bakedOverrides = new ContainedFluidOverrideHandler(new class_806(modelBaker, List.of()), modelBaker, List.of(), this);
        class_1087 bakedModel = new UniversalBucketBakedModel(quads, method_3443(), bakedOverrides, particleSprite);
        bakedOverrides.setBaseModel(bakedModel);

        return bakedModel;
    }

    private boolean isValid(class_4730 material) {
        return material != null && !material.method_24147().equals(class_1047.method_4539());
    }

    private static final class ContainedFluidOverrideHandler extends class_806 {

        private final Map<String, class_1087> cache = Maps.newHashMap(); // contains all the baked models since they'll never change
        private final class_806 nested;
        private class_1087 baseModel = null;
        private final class_7775 baker;
        private final UniversalBucketUnbakedModel parent;

        private Integer upperBreakTemperature = null;
        private Integer lowerBreakTemperature = null;

        private ContainedFluidOverrideHandler(class_806 nested, class_7775 baker, List<class_799> itemOverrides, UniversalBucketUnbakedModel parent) {
            super(baker, itemOverrides);
            this.nested = nested;
            this.baker = baker;
            this.parent = parent;
        }

        private void setBaseModel(class_1087 baseModel) {
            this.baseModel = baseModel;
        }

        @Nullable
        @Override
        public class_1087 method_64063(@Nonnull class_1799 stack, @Nullable class_638 level, @Nullable class_1309 entity, int seed) {
            class_1087 overridden = nested.method_64063(stack, level, entity, seed);
            if (overridden != null) return overridden;
            if (stack.method_7909() instanceof UniversalBucketItem bucket) {
                boolean containsEntityType = false;
                String content = BucketLibUtil.getEntityTypeString(stack);
                if (content != null) {
                    containsEntityType = true;
                } else {
                    content = BucketLibUtil.getContentString(stack);
                }
                class_3611 fluid = null;
                if (content == null) {
                    fluid = BucketLibUtil.getFluid(stack);
                    class_2960 location = class_7923.field_41173.method_10221(fluid);
                    content = (location != class_7923.field_41173.method_10137()) ? location.toString() : null;
                }
                //reset cache if temperature config changed
                if (!Objects.equals(upperBreakTemperature, bucket.getUpperBreakTemperature()) || !Objects.equals(lowerBreakTemperature, bucket.getLowerBreakTemperature())) {
                    upperBreakTemperature = bucket.getUpperBreakTemperature();
                    lowerBreakTemperature = bucket.getLowerBreakTemperature();
                    cache.clear();
                }
                class_1087 bakedModel = cache.get(content);
                if (bakedModel == null && content != null) {
                    boolean isCracked = bucket.isCracked(stack);
                    UniversalBucketUnbakedModel unbaked = (fluid == null) ? this.parent.withOtherContent(class_2960.method_60654(content), isCracked, containsEntityType) : this.parent.withFluid(fluid, isCracked);
                    bakedModel = unbaked.method_4753(baker, class_4730::method_24148, class_1086.field_5350);
                    cache.put(content, bakedModel);
                }
                return bakedModel;
            }
            return baseModel;
        }
    }

}

==> bucketlib-fabric-1.21.3-4.2.4.0-sources.jar: de/cech12/bucketlib/client/model/UniversalBucketUnbakedModel.java <==
package de.cech12.bucketlib.client.model;

import com.google.common.collect.Maps;
import de.cech12.bucketlib.api.BucketLib;
import de.cech12.bucketlib.api.BucketLibTags;
import de.cech12.bucketlib.api.item.UniversalBucketItem;
import de.cech12.bucketlib.client.BucketLibClientMod;
import de.cech12.bucketlib.mixin.BlockModelAccessor;
import de.cech12.bucketlib.util.BucketLibUtil;
import net.fabricmc.fabric.api.client.render.fluid.v1.FluidRenderHandlerRegistry;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariantAttributes;
import net.minecraft.class_1047;
import net.minecraft.class_1058;
import net.minecraft.class_1086;
import net.minecraft.class_1087;
import net.minecraft.class_1100;
import net.minecraft.class_1309;
import net.minecraft.class_1723;
import net.minecraft.class_1799;
import net.minecraft.class_2960;
import net.minecraft.class_3611;
import net.minecraft.class_3612;
import net.minecraft.class_3665;
import net.minecraft.class_4730;
import net.minecraft.class_638;
import net.minecraft.class_777;
import net.minecraft.class_7775;
import net.minecraft.class_7923;
import net.minecraft.class_793;
import net.minecraft.class_799;
import net.minecraft.class_806;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;

/**
 * This implementation is based on net.neoforged.neoforge.client.model.DynamicFluidContainerModel.
 * Multiple changes were done to simplify the class
 */
public class UniversalBucketUnbakedModel extends class_793 implements class_1100 {

    private static final Map<class_2960, class_2960> TEXTURE_MAP = Maps.newHashMap();
    private static final class_4730 MISSING_LOWER_CONTENT_MATERIAL = new class_4730(class_1723.field_21668, getContentTexture(BucketLib.id("missing_lower_content")));

    @Nonnull
    private class_3611 fluid = class_3612.field_15906;
    @Nullable
    private class_2960 otherContent = null;

    private boolean isCracked = false;
    private boolean isLower = false;

    public UniversalBucketUnbakedModel(class_793 blockModel) {
        super(
                ((BlockModelAccessor) blockModel).bucketlib_getParentLocation(),
                blockModel.method_3433(),
                ((BlockModelAccessor) blockModel).bucketlib_getTextureMap(),
                blockModel.method_3444(),
                blockModel.method_24298(),
                blockModel.method_3443(),
                blockModel.method_3434()
        );
    }

    public UniversalBucketUnbakedModel(class_793 blockModel, @Nonnull class_3611 fluid, @Nullable class_2960 otherContent, boolean isCracked, boolean isLower) {
        this(blockModel);
        this.fluid = fluid;
        this.otherContent = otherContent;
        this.isCracked = isCracked;
        this.isLower = isLower;
    }

    /**
     * Returns a new UniversalBucketModel with the given fluid.
     */
    public UniversalBucketUnbakedModel withFluid(class_3611 newFluid, boolean isCracked) {
        return new UniversalBucketUnbakedModel(this, newFluid, null, isCracked, false);
    }

    /**
     * Returns a new UniversalBucketModel with the given other content.
     */
    public UniversalBucketUnbakedModel withOtherContent(class_2960 otherContent, boolean isCracked, boolean isLower) {
        return new UniversalBucketUnbakedModel(this, class_3612.field_15906, otherContent, isCracked, isLower);
    }

    private static class_2960 getContentTexture(class_2960 otherContentLocation) {
        class_2960 texture = TEXTURE_MAP.get(otherContentLocation);
        if (texture == null) {
            String textureLocation = String.format("item/bucket_content/%s", otherContentLocation.method_12832());
            texture = otherContentLocation.method_45136(textureLocation);
            TEXTURE_MAP.put(otherContentLocation, texture);
        }
        return texture;
    }

    public class_2960 getParentLocation() {
        return this.field_4247;
    }

    @Nonnull
    @Override
    public class_793 method_3431() {
        if (this.field_4253 == null) {
            return this;
        }
        if (BucketLibClientMod.UNIVERSAL_BUCKET_MODEL.equals(this.getParentLocation())) {
            //skip the BlockModel swap of the vanilla ItemRendering (ItemModelGenerator)
            return this.field_4253;
        }
        return this.field_4253.method_3431();
    }

    @Nonnull
    @Override
    public class_1087 method_4753(@Nonnull class_7775 modelBaker, @Nonnull Function<class_4730, class_1058> spriteGetter, @Nonnull class_3665 modelState) {
        //resolve parents to use their defined materials, too
        if (this.field_4253 == null) {
            method_62326(BucketLibClientMod::getModel);
        }
        class_4730 particleLocation = this.method_24077("particle");
        class_1058 particleSprite = null;
        if (isValid(particleLocation, spriteGetter)) {
            particleSprite = spriteGetter.apply(particleLocation);
        }

        class_4730 baseLocation = null;
        if (this.isLower) {
            if (this.isCracked) {
                baseLocation = this.method_24077("crackedLowerBase");
            }
            if (!isValid(baseLocation, spriteGetter)) {
                baseLocation = this.method_24077("lowerBase");
            }
        } else {
            if (this.isCracked) {
                baseLocation = this.method_24077("crackedBase");
            }
            if (!isValid(baseLocation, spriteGetter)) {
                baseLocation = this.method_24077("base");
            }
        }

        class_4730 otherContentLocation = null;
        class_4730 fluidLocation = null;
        class_4730 fluidMaskLocation = null;
        if (this.otherContent != null) {
            otherContentLocation = new class_4730(class_1723.field_21668, getContentTexture(this.otherContent));
        }
        if (this.fluid != class_3612.field_15906) {
            fluidLocation = new class_4730(class_1723.field_21668, getContentTexture(class_7923.field_41173.method_10221(this.fluid)));
            if (this.isCracked) {
                fluidMaskLocation = this.method_24077("crackedFluidMask");
            }
            if (!isValid(fluidMaskLocation, spriteGetter)) {
                fluidMaskLocation = this.method_24077("fluidMask");
            }
        }
        //oversteer fluid texture if available
        if (otherContent == null && isValid(fluidLocation, spriteGetter)) {
            otherContentLocation = fluidLocation;
        }

        class_1058 baseSprite = null;
        if (isValid(baseLocation, spriteGetter)) {
            baseSprite = spriteGetter.apply(baseLocation);
            if (particleSprite == null)
                particleSprite = baseSprite;
        }
        class_1058 otherContentSprite;
        if (isValid(otherContentLocation, spriteGetter)) {
            otherContentSprite = spriteGetter.apply(otherContentLocation);
            if (particleSprite == null)
                particleSprite = otherContentSprite;
        } else {
            //if content texture is missing - fallback to pink content texture>
            otherContentSprite = spriteGetter.apply(MISSING_LOWER_CONTENT_MATERIAL);
        }

        class_1058 fluidSprite = null;
        if (fluid != class_3612.field_15906 && FluidRenderHandlerRegistry.INSTANCE.get(fluid) != null) {
            class_1058[] sprites = Objects.requireNonNull(FluidRenderHandlerRegistry.INSTANCE.get(fluid)).getFluidSprites(null, null, fluid.method_15785());
            if (sprites.length > 0) {
                fluidSprite = sprites[0];
                if (particleSprite == null)
                    particleSprite = fluidSprite;
            }
        }

        // if the fluid is lighter than air, will manipulate the initial state to be rotated 180deg to turn it upside down
        if (fluid != class_3612.field_15906 && !fluid.method_15785().method_15767(BucketLibTags.Fluids.NO_FLIPPING) && FluidVariantAttributes.isLighterThanAir(FluidVariant.of(fluid))) {
            modelState = class_1086.field_5358;
        }

        List<class_777> quads = new ArrayList<>();

        if (baseSprite != null) {
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemElements(0, "base", baseSprite.method_45851()),
                    baseSprite, modelState));
        }
        if (isValid(otherContentLocation, spriteGetter)) {
            assert otherContentSprite != null;
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemElements(-1, "content", otherContentSprite.method_45851()),
                    otherContentSprite, modelState));
        } else if (isValid(fluidMaskLocation, spriteGetter) && fluid != class_3612.field_15906) {
            class_1058 templateSprite = spriteGetter.apply(fluidMaskLocation);
            quads.addAll(GeometryUtils.bakeElements(this,
                    GeometryUtils.createUnbakedItemMaskElements(1, "fluid", templateSprite.method_45851()),
                    fluidSprite, modelState));
        }

        ContainedFluidOverrideHandler bakedOverrides = new ContainedFluidOverrideHandler(new class_806(modelBaker, List.of()), modelBaker, List.of(), this);
        class_1087 bakedModel = new UniversalBucketBakedModel(quads, method_3443(), bakedOverrides, particleSprite);
        bakedOverrides.setBaseModel(bakedModel);

        return bakedModel;
    }

    private boolean isValid(class_4730 material, Function<class_4730, class_1058> spriteGetter) {
        return material != null && !class_1047.method_4539().equals(spriteGetter.apply(material).method_45851().method_45816());
    }

    private static final class ContainedFluidOverrideHandler extends class_806 {

        private final Map<String, class_1087> cache = Maps.newHashMap(); // contains all the baked models since they'll never change
        private final class_806 nested;
        private class_1087 baseModel = null;
        private final class_7775 baker;
        private final UniversalBucketUnbakedModel parent;

        private Integer upperBreakTemperature = null;
        private Integer lowerBreakTemperature = null;

        private ContainedFluidOverrideHandler(class_806 nested, class_7775 baker, List<class_799> itemOverrides, UniversalBucketUnbakedModel parent) {
            super(baker, itemOverrides);
            this.nested = nested;
            this.baker = baker;
            this.parent = parent;
        }

        private void setBaseModel(class_1087 baseModel) {
            this.baseModel = baseModel;
        }

        @Nullable
        @Override
        public class_1087 method_64063(@Nonnull class_1799 stack, @Nullable class_638 level, @Nullable class_1309 entity, int seed) {
            class_1087 overridden = nested.method_64063(stack, level, entity, seed);
            if (overridden != null) return overridden;
            if (stack.method_7909() instanceof UniversalBucketItem bucket) {
                boolean containsEntityType = false;
                String content = BucketLibUtil.getEntityTypeString(stack);
                if (content != null) {
                    containsEntityType = true;
                } else {
                    content = BucketLibUtil.getContentString(stack);
                }
                class_3611 fluid = null;
                if (content == null) {
                    fluid = BucketLibUtil.getFluid(stack);
                    class_2960 location = class_7923.field_41173.method_10221(fluid);
                    content = (location != class_7923.field_41173.method_10137()) ? location.toString() : null;
                }
                //reset cache if temperature config changed
                if (!Objects.equals(upperBreakTemperature, bucket.getUpperBreakTemperature()) || !Objects.equals(lowerBreakTemperature, bucket.getLowerBreakTemperature())) {
                    upperBreakTemperature = bucket.getUpperBreakTemperature();
                    lowerBreakTemperature = bucket.getLowerBreakTemperature();
                    cache.clear();
                }
                class_1087 bakedModel = cache.get(content);
                if (bakedModel == null && content != null) {
                    boolean isCracked = bucket.isCracked(stack);
                    UniversalBucketUnbakedModel unbaked = (fluid == null) ? this.parent.withOtherContent(class_2960.method_60654(content), isCracked, containsEntityType) : this.parent.withFluid(fluid, isCracked);
                    bakedModel = unbaked.method_4753(baker, class_4730::method_24148, class_1086.field_5350);
                    cache.put(content, bakedModel);
                }
                return bakedModel;
            }
            return baseModel;
        }
    }

}
